/*-
 * For Copyright information see RFC7541 [BSD3]
 */

STAT_HDRS(1,   ":authority",                   "")
STAT_HDRS(2,   ":method",                      "GET")
STAT_HDRS(3,   ":method",                      "POST")
STAT_HDRS(4,   ":path",                        "/")
STAT_HDRS(5,   ":path",                        "/index.html")
STAT_HDRS(6,   ":scheme",                      "http")
STAT_HDRS(7,   ":scheme",                      "https")
STAT_HDRS(8,   ":status",                      "200")
STAT_HDRS(9,   ":status",                      "204")
STAT_HDRS(10,  ":status",                      "206")
STAT_HDRS(11,  ":status",                      "304")
STAT_HDRS(12,  ":status",                      "400")
STAT_HDRS(13,  ":status",                      "404")
STAT_HDRS(14,  ":status",                      "500")
STAT_HDRS(15,  "accept-charset",               "")
STAT_HDRS(16,  "accept-encoding",              "gzip,deflate")
STAT_HDRS(17,  "accept-language",              "")
STAT_HDRS(18,  "accept-ranges",                "")
STAT_HDRS(19,  "accept",                       "")
STAT_HDRS(20,  "access-control-allow-origin",  "")
STAT_HDRS(21,  "age",                          "")
STAT_HDRS(22,  "allow",                        "")
STAT_HDRS(23,  "authorization",                "")
STAT_HDRS(24,  "cache-control",                "")
STAT_HDRS(25,  "content-disposition",          "")
STAT_HDRS(26,  "content-encoding",             "")
STAT_HDRS(27,  "content-language",             "")
STAT_HDRS(28,  "content-length",               "")
STAT_HDRS(29,  "content-location",             "")
STAT_HDRS(30,  "content-range",                "")
STAT_HDRS(31,  "content-type",                 "")
STAT_HDRS(32,  "cookie",                       "")
STAT_HDRS(33,  "date",                         "")
STAT_HDRS(34,  "etag",                         "")
STAT_HDRS(35,  "expect",                       "")
STAT_HDRS(36,  "expires",                      "")
STAT_HDRS(37,  "from",                         "")
STAT_HDRS(38,  "host",                         "")
STAT_HDRS(39,  "if-match",                     "")
STAT_HDRS(40,  "if-modified-since",            "")
STAT_HDRS(41,  "if-none-match",                "")
STAT_HDRS(42,  "if-range",                     "")
STAT_HDRS(43,  "if-unmodified-since",          "")
STAT_HDRS(44,  "last-modified",                "")
STAT_HDRS(45,  "link",                         "")
STAT_HDRS(46,  "location",                     "")
STAT_HDRS(47,  "max-forwards",                 "")
STAT_HDRS(48,  "proxy-authenticate",           "")
STAT_HDRS(49,  "proxy-authorization",          "")
STAT_HDRS(50,  "range",                        "")
STAT_HDRS(51,  "referer",                      "")
STAT_HDRS(52,  "refresh",                      "")
STAT_HDRS(53,  "retry-after",                  "")
STAT_HDRS(54,  "server",                       "")
STAT_HDRS(55,  "set-cookie",                   "")
STAT_HDRS(56,  "strict-transport-security",    "")
STAT_HDRS(57,  "transfer-encoding",            "")
STAT_HDRS(58,  "user-agent",                   "")
STAT_HDRS(59,  "vary",                         "")
STAT_HDRS(60,  "via",                          "")
STAT_HDRS(61,  "www-authenticate",             "")
